#include "ble_component.h"

static const uint8_t E2_NAMESPACE_IN_DECIMAL[NAMESPACE_SIZE] = {170, 170, 187, 187, 204, 204, 221, 221, 238, 238};

static const uint8_t HUB_NAMESPACE_IN_DECIMAL[NAMESPACE_SIZE] = {170, 170, 187, 187, 204, 204, 221, 221, 170, 170};

static const uint8_t TAG_NAMESPACE_IN_DECIMAL[NAMESPACE_SIZE] = {170, 170, 187, 187, 204, 204, 221, 221, 187, 187};

static int current_pointer_index_ = -1; // This will point to the current index, where the latest location data will be saved.

static int starting_index_ = -1;

static tag_data tag_data_array_[MAX_DATA_QUEUE_LEN];

static ble_evt_handler_t app_ble_evt_handler_;

static ble_gap_addr_t *tag_mac_addr_p = NULL;

static uint32_t *tag_start_epoch_p;

static uint64_t scan_finished_epoch;

static bool is_data_transfer_in_progress_;

static const uint8_t MAX_DATA_ITEMS_IN_PACKET = 4;

static bool is_gatt_server_running_;

BLE_CUS_DEF(m_cus);

#define CLEAR_DATA_QUEUE() (memset(tag_data_array_, 0, sizeof(tag_data) * MAX_DATA_QUEUE_LEN));

NRF_BLE_SCAN_DEF(m_scan);           /**< Scanning module instance. */
NRF_BLE_GATT_DEF(m_gatt);           /**< GATT module instance. */
BLE_ADVERTISING_DEF(m_advertising); /**< Advertising module instance. */

/**< Scan parameters requested for scanning and connection. */
static ble_gap_scan_params_t const m_scan_param = {
    .active = 0x01,
    .interval = NRF_BLE_SCAN_SCAN_INTERVAL,
    .window = NRF_BLE_SCAN_SCAN_WINDOW,
    .filter_policy = BLE_GAP_SCAN_FP_ACCEPT_ALL, // BLE_GAP_SCAN_FP_WHITELIST,
    .timeout = NRF_BLE_SCAN_SCAN_DURATION,
    .scan_phys = BLE_GAP_PHY_1MBPS,
};

#if TAG_DEBUG_MODE
void print_hub_beacon_data(eddy_stone_adv_data *adv_data) {
  char log_buffer[150];
  sprintf(log_buffer, "UUID:%02x%02x,Type:%02x,Rx:%02x,Address:%02x:%02x:%02x:%02x:%02x:%02x,NamespaceId:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x,InstanceId:%02x:%02x:%02x:%02x:%02x:%02x\0", adv_data->uuid[0], adv_data->uuid[1],
      adv_data->type, adv_data->rssi, adv_data->address[5], adv_data->address[4], adv_data->address[3], adv_data->address[2], adv_data->address[1], adv_data->address[0], adv_data->namespace_id[0], adv_data->namespace_id[1], adv_data->namespace_id[2],
      adv_data->namespace_id[3], adv_data->namespace_id[4], adv_data->namespace_id[5], adv_data->namespace_id[6], adv_data->namespace_id[7], adv_data->namespace_id[8], adv_data->namespace_id[9], adv_data->instance_id[0],
      adv_data->instance_id[1], adv_data->instance_id[2], adv_data->instance_id[3], adv_data->instance_id[4], adv_data->instance_id[5]);
  NRF_LOG_DEBUG("BLE_COMPONENT-- Beacon data: %s", nrf_log_push(log_buffer));
}
#endif

/**
 * @Brief This Function is to parse the eddy stone hub beacon.
 **/
static void parse_hub_beacon_to_eddy_stone_format(const ble_gap_evt_adv_report_t *p_adv_report, eddy_stone_adv_data *adv_data) {
  uint16_t offset = 0;
  uint16_t length = ble_advdata_search(p_adv_report->data.p_data, p_adv_report->data.len, &offset, 0x16);

  if (length != 0) {
    memcpy(adv_data->uuid, &p_adv_report->data.p_data[offset], HUB_BEACON_UUID_LEN); // eddystone UUID
    if (adv_data->uuid[0] != EDDY_STONE_UUID_HSB || adv_data->uuid[1] != EDDY_STONE_UUID_LSB) {
      return;
    }
    adv_data->type = p_adv_report->data.p_data[offset + 2];                                             // Eddy stone beacon Type
    adv_data->rx_power = p_adv_report->data.p_data[offset + 3];                                         // Rx value
    memcpy(adv_data->namespace_id, &p_adv_report->data.p_data[offset + 4], HUB_BEACON_NAMESPACE_LEN);   // namespace id 10 bytes
    memcpy(adv_data->instance_id, &p_adv_report->data.p_data[offset + 14], HUB_BEACON_INSTANCE_ID_LEN); // instance id 6 bytes
    adv_data->rssi = p_adv_report->rssi;                                                                // rssi
    memcpy(adv_data->address, &p_adv_report->peer_addr.addr, HUB_BEACON_MAC_ADDRESS_LEN);               // address
  }                                                                                                     // endif
}

void stop_gatt_server() {
  NRF_LOG_DEBUG("BLE_COMPONENT-- Stopping the GATT Server!");
  is_gatt_server_running_ = false;
  stop_gatt_service();
  is_data_transfer_in_progress_ = false;
}

static void terminate_gatt_server() {
  NRF_LOG_DEBUG("BLE_COMPONENT-- Terminating the GATT Server!");
  current_pointer_index_ = -1;
  starting_index_ = -1;
  CLEAR_DATA_QUEUE(); // Clearing the data queue
  stop_gatt_server();
}

/**@brief Function handle the BLE beacon scanning results.
 */
static void scan_evt_handler(scan_evt_t const *p_scan_evt) {
  if (p_scan_evt->scan_evt_id == NRF_BLE_SCAN_EVT_SCAN_TIMEOUT) {
    scan_finished_epoch = millis();
    NRF_LOG_INFO("BLE_COMPONENT-- Scan timed out at time: %lld", scan_finished_epoch);
    return;
  } // end if

  register const ble_gap_evt_adv_report_t *p_adv_report = p_scan_evt->params.filter_match.p_adv_report;

  eddy_stone_adv_data *adv_data_p = (eddy_stone_adv_data *)malloc(sizeof(eddy_stone_adv_data));

  if (adv_data_p == NULL) {
    NRF_LOG_ERROR("BLE_COMPONENT-- No memory to create the adv data!");
    sd_nvic_SystemReset();
  } // end if

  parse_hub_beacon_to_eddy_stone_format(p_adv_report, adv_data_p);

  if (adv_data_p->uuid[0] == EDDY_STONE_UUID_HSB && adv_data_p->uuid[1] == EDDY_STONE_UUID_LSB) {
#if TAG_DEBUG_MODE
    print_hub_beacon_data(adv_data_p);
#endif
    if (memcmp(adv_data_p->namespace_id, E2_NAMESPACE_IN_DECIMAL, HUB_BEACON_NAMESPACE_LEN) == 0) {
      if (current_pointer_index_ >= MAX_DATA_QUEUE_LEN || current_pointer_index_ < 0) {
        NRF_LOG_DEBUG("BLE_COMPONENT-- Resetting the counter");
        current_pointer_index_ = 0;
      } // end if

      tag_data *tag_data_p = (tag_data *)malloc(sizeof(tag_data));
      if (tag_data_p == NULL) {
        NRF_LOG_ERROR("BLE_COMPONENT-- No memory to create the tag data!");
        sd_nvic_SystemReset();
      } // end if

      tag_data_p->contact_mS = millis();
      memcpy(tag_data_p->address, adv_data_p->address, sizeof(adv_data_p->address));
      tag_data_p->rssi = adv_data_p->rssi;
      tag_data_p->battery_percentage = get_battery_level();
      tag_data_array_[current_pointer_index_] = *tag_data_p;
            nrf_delay_ms(10);//cool down
      NRF_LOG_DEBUG("BLE_COMPONENT-- Saved the data at location at %d contact time: %lld", current_pointer_index_, tag_data_array_[current_pointer_index_].contact_mS);
      if (tag_data_p != NULL) {
        free(tag_data_p);
      } // end if
      current_pointer_index_++;
#if TAG_DEBUG_MODE
      switch_led_debug_mode(TAG_RECEIVING_E2_DATA);
#endif
    } else if (memcmp(adv_data_p->namespace_id, HUB_NAMESPACE_IN_DECIMAL, HUB_BEACON_NAMESPACE_LEN) == 0) {
      if (current_pointer_index_ < 4) {
        NRF_LOG_INFO("BLE_COMPONENT-- Hub beacon detected, but no data to share, keep on adverstising....");
        if (adv_data_p != NULL) {
          free(adv_data_p);
        } // end if
        return;
      }
      NRF_LOG_INFO("BLE_COMPONENT-- Hub beacon detected, switching to gatt server mode  for data transfer starting from index %d", current_pointer_index_);
      start_gatt_server();
    } else {
      // Do nothing
    }
  } else {
    NRF_LOG_DEBUG("BLE_COMPONENT-- Other ble adv type dectected....");
  }
  if (adv_data_p != NULL) {
    free(adv_data_p);
  } // end if
}

/**@brief Function for initializing the BLE stack.

 @details Initializes the SoftDevice and the BLE event interrupt.
*/
static void ble_stack_init(void) {
  ret_code_t err_code;

  err_code = nrf_sdh_enable_request();
  APP_ERROR_CHECK(err_code);

  // Configure the BLE stack using the default settings.
  // Fetch the start address of the application RAM.
  uint32_t ram_start = 0;
  err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
  APP_ERROR_CHECK(err_code);

  // Enable BLE stack.
  err_code = nrf_sdh_ble_enable(&ram_start);
  APP_ERROR_CHECK(err_code);
  CLEAR_DATA_QUEUE();
}

/**@brief Function for the GAP initialization.


   @details This function sets up all the necessary GAP (Generic Access Profile) parameters of the
            device including the device name, appearance, and the preferred connection parameters.
 */
static void gap_params_init(void) {
  ret_code_t err_code;
  ble_gap_conn_params_t gap_conn_params;
  ble_gap_conn_sec_mode_t sec_mode;

  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

  err_code = sd_ble_gap_device_name_set(&sec_mode,
      (const uint8_t *)DEVICE_NAME,
      strlen(DEVICE_NAME));
  APP_ERROR_CHECK(err_code);

  memset(&gap_conn_params, 0, sizeof(gap_conn_params));

  gap_conn_params.min_conn_interval = MSEC_TO_UNITS(MIN_CONN_INTERVAL_MSEC, UNIT_1_25_MS);
  gap_conn_params.max_conn_interval = MSEC_TO_UNITS(MAX_CONN_INTERVAL_MSEC, UNIT_1_25_MS);
  gap_conn_params.slave_latency = SLAVE_LATENCY;
  gap_conn_params.conn_sup_timeout = MSEC_TO_UNITS(CONN_SUP_TIMEOUT_MSEC, UNIT_10_MS);

  err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
  APP_ERROR_CHECK(err_code);
}

static void start_data_transfer(ble_cus_t *p_cus) {
  if (current_pointer_index_ == -1) {
    NRF_LOG_DEBUG("Stopping the data tranfer as there is no data to share");
    terminate_gatt_server();
    return;
  } else if (starting_index_ == -1) {
    current_pointer_index_--;
    NRF_LOG_DEBUG("Starting Data Tranfer sending data as position %d", current_pointer_index_);
    starting_index_ = current_pointer_index_;
  } else if (starting_index_ == current_pointer_index_) {
    NRF_LOG_DEBUG("Stopping the data tranfer with start index: %d", starting_index_);
    terminate_gatt_server();
    return;
  } else if (current_pointer_index_ < 0) {
    current_pointer_index_ = MAX_DATA_QUEUE_LEN - 1;
  } else {
    // DO nothing
  }
  NRF_LOG_DEBUG("Sending data item from position %d", current_pointer_index_);
  if (tag_data_array_[current_pointer_index_].contact_mS == 0) {
    NRF_LOG_DEBUG("Stopping the data tranfer with start index: %d as no data captured at this index", starting_index_);
    terminate_gatt_server();
    return;
  } // end if

  char tag_packet_data[NRF_SDH_BLE_GATT_MAX_MTU_SIZE] = {0};
  int data_items_added = 0;
  int start_index = current_pointer_index_;
  is_data_transfer_in_progress_ = true;
  while (data_items_added < MAX_DATA_ITEMS_IN_PACKET) {
    char temp_tag_packet_data[NRF_SDH_BLE_GATT_MAX_MTU_SIZE / 4] = {0};
    uint64_t current_time = millis();
    if (tag_data_array_[start_index].contact_mS == 0) {
      NRF_LOG_DEBUG("No more data left to add into the packet");
      break;
    } // end if

    uint32_t delta_time = current_time - tag_data_array_[start_index].contact_mS;
    NRF_LOG_DEBUG("The delta time is %lld, with current time %lld and contact_time %lld", delta_time, current_time, tag_data_array_[start_index].contact_mS);
    sprintf(temp_tag_packet_data, "%04x-%02x|%04x|%02x:%02x:%02x:%02x:%02x:%02x|%02x:%02x:%02x:%02x:%02x:%02x|%08x\0", start_index, tag_data_array_[start_index].rssi, tag_data_array_[start_index].battery_percentage, tag_data_array_[start_index].address[5],
        tag_data_array_[start_index].address[4], tag_data_array_[start_index].address[3], tag_data_array_[start_index].address[2], tag_data_array_[start_index].address[1], tag_data_array_[start_index].address[0], tag_mac_addr_p->addr[5], tag_mac_addr_p->addr[4],
        tag_mac_addr_p->addr[3], tag_mac_addr_p->addr[2], tag_mac_addr_p->addr[1], tag_mac_addr_p->addr[0], delta_time);
    if (data_items_added > 0 && data_items_added < MAX_DATA_ITEMS_IN_PACKET) {
      strcat(tag_packet_data, ",");
    } // end if
    strcat(tag_packet_data, temp_tag_packet_data);
    data_items_added++;
    if (start_index > 0) {
      start_index--;
    } else {
      break;
    }
  } // end loop

  ret_code_t res_code = send_data_to_hub(p_cus, 
  );
  if (res_code != NRF_SUCCESS) {
    NRF_LOG_ERROR("BLE_COMPONENT-- Data Packet transfer failed %d of data %s. \r\n ", res_code, tag_packet_data);
    stop_gatt_server();
  } else {
    NRF_LOG_INFO("BLE_COMPONENT-- Data Packet transfer success ", tag_packet_data);
    if (current_pointer_index_ > 0) {
      current_pointer_index_ = current_pointer_index_ - MAX_DATA_ITEMS_IN_PACKET;
    } else if (current_pointer_index_ == 0) {
      current_pointer_index_ = -2;
    } else {
      // Do NOthing
    }
    memset(&(tag_data_array_[current_pointer_index_]), 0, sizeof(tag_data));

#if TAG_DEBUG_MODE
    switch_led_debug_mode(TAG_TO_HUB_TRANSFER_COMPLETE);
#endif
  }
}

/**@brief Function for handling the Write event.
 *
 * @param[in]   p_cus       Custom Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_write(ble_cus_t *p_cus, ble_evt_t const *p_ble_evt) {
  ble_gatts_evt_write_t const *p_evt_write = &p_ble_evt->evt.gatts_evt.params.write;

  NRF_LOG_DEBUG("BLE_COMPONENT-- Write Event Handler, the data is %02x", p_evt_write->data);
  int16_t write_mode = -1;

  memcpy(&write_mode, p_evt_write->data, p_evt_write->len);

  if (write_mode == 2) {
    NRF_LOG_DEBUG("BLE_COMPONENT-- Starting the data tranfer");
    start_data_transfer(p_cus);
  } else if (write_mode == -1) {
    terminate_gatt_server();
  } else {
    NRF_LOG_DEBUG("BLE_COMPONENT-- Invalid write mode ! %d", write_mode);
    return;
  }
}

/**@brief Function for handling the indicate event.
 *
 * @param[in]   p_cus       Custom Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_indicate(ble_cus_t *p_cus, ble_evt_t const *p_ble_evt) {
  NRF_LOG_DEBUG("BLE_COMPONENT-- Indicate Event Handler");
  start_data_transfer(p_cus);
  nrf_delay_ms(10);
}

/**@brief Function for handling the indicate event.
 *
 * @param[in]   p_cus       Custom Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_read(ble_cus_t *p_cus, ble_evt_t const *p_ble_evt) {
  NRF_LOG_DEBUG("BLE_COMPONENT-- Read Event Handler");
}

/**@brief Function for initializing services that will be used by the application.
 */
static void services_init() {
  ret_code_t err_code;

  ble_cus_init_t cus_init = {0};

  // Initialize CUS Service init structure to zero.
  cus_init.evt_handler = app_ble_evt_handler_;

  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cus_init.custom_value_char_attr_md.cccd_write_perm);
  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cus_init.custom_value_char_attr_md.read_perm);
  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cus_init.custom_value_char_attr_md.write_perm);

  err_code = gatt_services_init(&m_cus, &cus_init, on_write, on_indicate, on_read);
  APP_ERROR_CHECK(err_code);
}

void start_gatt_server() {
  NRF_LOG_DEBUG("Starting the GATT Server!");
  is_gatt_server_running_ = true;
  stop_hub_beacon_scan();
  gatts_advertising_start();
}

void gatt_module_init(const ble_evt_handler_t ble_handler, const ble_adv_evt_handler_t ble_adv_evt_handler, ble_gap_addr_t *tag_mac_addr, uint32_t *tag_start_epoch) {
  NRF_LOG_DEBUG("GATT--Init the GATT Module");
  current_pointer_index_ = -1;

  app_ble_evt_handler_ = ble_handler;

  tag_mac_addr_p = tag_mac_addr;

  tag_start_epoch_p = tag_start_epoch;

  gap_params_init();
  APP_ERROR_CHECK(nrf_ble_gatt_init(&m_gatt, NULL));
  services_init();
  gatts_advertising_init(ble_adv_evt_handler);
  CLEAR_DATA_QUEUE();
}

void gatts_advertising_init(const ble_adv_evt_handler_t ble_adv_evt_handler) {
  NRF_LOG_DEBUG("BLE_COMPONENT-- Init the GATTS advertisement");
  ret_code_t err_code;
  ble_advertising_init_t init;

  memset(&init, 0, sizeof(init));

  init.advdata.name_type = BLE_ADVDATA_FULL_NAME;
  init.advdata.include_appearance = true;
  init.advdata.flags = BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE;
  init.config.ble_adv_fast_enabled = true;
  init.config.ble_adv_fast_interval = ADV_INTERVAL;
  init.config.ble_adv_fast_timeout = APP_ADV_DURATION;
  init.evt_handler = ble_adv_evt_handler;

  err_code = ble_advertising_init(&m_advertising, &init);
  APP_ERROR_CHECK(err_code);
  ble_advertising_conn_cfg_tag_set(&m_advertising, APP_BLE_CONN_CFG_TAG);
}

void gatts_advertising_start() {
  NRF_LOG_DEBUG("BLE COMP--Starting the GATTS advertisement");
  APP_ERROR_CHECK(ble_advertising_start(&m_advertising, BLE_ADV_MODE_FAST));
}

void stop_hub_beacon_scan(void) {
  NRF_LOG_DEBUG("BLE_COMPONENT-- Stopping the Edddy Stone Beacon Scan");
  nrf_ble_scan_stop();
}

void start_hub_beacon_scan(void) {
  NRF_LOG_DEBUG("BLE_COMPONENT-- Starting the Edddy Stone Beacon Scan");
  if (!is_gatt_server_running_) {
    scan_finished_epoch = 0;
    APP_ERROR_CHECK(nrf_ble_scan_start(&m_scan));
  }
}

void hub_beacon_scan_init() {
  NRF_LOG_DEBUG("BLE_COMPONENT-- Initialsing the HUB scanning");
  ble_stack_init();
  ret_code_t err_code;
  nrf_ble_scan_init_t init_scan;
  memset(&init_scan, 0, sizeof(init_scan));
  init_scan.p_scan_param = &m_scan_param;
  init_scan.connect_if_match = false;
  init_scan.conn_cfg_tag = APP_BLE_CONN_CFG_TAG;
  err_code = nrf_ble_scan_init(&m_scan, &init_scan, &scan_evt_handler);
  APP_ERROR_CHECK(err_code);
  CLEAR_DATA_QUEUE();
  is_gatt_server_running_ = false;
}

uint64_t get_scan_finshed_epoch() {
  return scan_finished_epoch;
}

bool is_data_transfer_in_progress() {
  return is_data_transfer_in_progress_;
}