#include "ble_service.h"

static tag_data *tag_data_array_p_;

static char tag_packet_data[DATA_PACKET_LENGTH];

static int *current_pointer_index_p;

static ble_gap_addr_t device_addr_;


/**@brief Function for handling the Connect event.
 *
 * @param[in]   p_cus       Custom Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_connect(ble_cus_t *p_cus, ble_evt_t const *p_ble_evt) {
  p_cus->conn_handle = p_ble_evt->evt.gap_evt.conn_handle;

  ble_cus_evt_t evt;

  evt.evt_type = BLE_CUS_EVT_CONNECTED;

  p_cus->evt_handler(p_cus, &evt);
}

/**@brief Function for handling the Disconnect event.
 *
 * @param[in]   p_cus       Custom Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_disconnect(ble_cus_t *p_cus, ble_evt_t const *p_ble_evt) {
  UNUSED_PARAMETER(p_ble_evt);
  p_cus->conn_handle = BLE_CONN_HANDLE_INVALID;

  ble_cus_evt_t evt;

  evt.evt_type = BLE_CUS_EVT_DISCONNECTED;

  p_cus->evt_handler(p_cus, &evt);
}

static void stop_gatt_service() {
  NRF_LOG_DEBUG("Stopping the GATT server and resetting all data");
  *current_pointer_index_p = -1;
  memset(tag_data_array_p_, 0, sizeof(tag_data) * MAX_DATA_QUEUE_LEN);
  // Disconnect all connected clients
  sd_ble_gap_disconnect(NULL, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
  // Disable GATT server
  sd_ble_gatts_service_changed(NULL, 0x0000, 0xFFFF);

}

static void update_tag_data() {
  register char *my_char_value_p = tag_packet_data;
  const int position = *current_pointer_index_p;

  sprintf(my_char_value_p, "%04x-%02x|%02x|%02x:%02x:%02x:%02x:%02x:%02x|%02x:%02x:%02x:%02x:%02x:%02x|%08x\0", position, tag_data_array_p_[position].rssi, tag_data_array_p_[position].battery_percentage, tag_data_array_p_[position].address[5],
      tag_data_array_p_[position].address[4], tag_data_array_p_[position].address[3], tag_data_array_p_[position].address[2], tag_data_array_p_[position].address[1], tag_data_array_p_[position].address[0], device_addr_.addr[5], device_addr_.addr[4],
      device_addr_.addr[3], device_addr_.addr[2], device_addr_.addr[1], device_addr_.addr[0], tag_data_array_p_[position].contact_mS);
}

static uint32_t ble_cus_custom_value_update(ble_cus_t *p_cus) {
  NRF_LOG_INFO("In ble_cus_custom_value_update. \r\n");
  if (p_cus == NULL) {
    return NRF_ERROR_NULL;
  }

  uint16_t data_len = strlen(tag_packet_data);
  ble_gatts_hvx_params_t hvx_params;
  memset(&hvx_params, 0, sizeof(hvx_params));
  hvx_params.handle = p_cus->custom_value_handles.value_handle;
  hvx_params.type = BLE_GATT_HVX_INDICATION;
  hvx_params.offset = 0;
  hvx_params.p_len = &data_len;
  hvx_params.p_data = (uint8_t *)tag_packet_data;

  ret_code_t err_code = sd_ble_gatts_hvx(p_cus->conn_handle, &hvx_params);

  if (err_code != NRF_SUCCESS) {
    NRF_LOG_INFO("Data tranfer failed %d. \r\n ", err_code);
    return err_code;
  }
  NRF_LOG_INFO("Data tranfer success %s",&(tag_data_array_p_[*current_pointer_index_p]));
  memset(&(tag_data_array_p_[*current_pointer_index_p]), 0, sizeof(tag_data));
  memset(tag_packet_data, 0, strlen(tag_packet_data));
#if TAG_DEBUG_MODE
  switch_led_debug_mode(TAG_GATT_CLIENT_CONNECTED);
#endif
  return err_code;
}

static void start_data_transfer(ble_cus_t *p_cus) {
  if (*current_pointer_index_p == -1) {
    NRF_LOG_DEBUG("Stopping the data tranfer as there is no data to share");
    stop_gatt_service();
    return;
  }//end if

  NRF_LOG_DEBUG("Sending data item from position %d", *current_pointer_index_p);
  update_tag_data();
  if (tag_data_array_p_[*current_pointer_index_p].contact_mS == 0) {
    NRF_LOG_DEBUG("Stopping the data tranfer as no data captured at this index");
    stop_gatt_service();
    return;
  } // end if
  ble_cus_custom_value_update(p_cus);
  if (*current_pointer_index_p > 0) {
    (*current_pointer_index_p)--;
  }
#if TAG_DEBUG_MODE
  switch_led_debug_mode(TAG_TO_HUB_TRANSFER_COMPLETE);
#endif
}

/**@brief Function for handling the Write event.
 *
 * @param[in]   p_cus       Custom Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_write(ble_cus_t *p_cus, ble_evt_t const *p_ble_evt) {
  ble_gatts_evt_write_t const *p_evt_write = &p_ble_evt->evt.gatts_evt.params.write;

  NRF_LOG_DEBUG("The data is %02x", p_evt_write->data);
  int16_t expected_index = -1;

  memcpy(&expected_index, p_evt_write->data, p_evt_write->len);

  if (expected_index == 1) {
    NRF_LOG_DEBUG("Starting the data tranfer");
    start_data_transfer(p_cus);
  } else if (expected_index == -1) {
    stop_gatt_service();
  } else {
    NRF_LOG_DEBUG("Invalid data range given! %d", expected_index);
    return;
  }
}

void ble_cus_on_ble_evt(ble_evt_t const *p_ble_evt, void *p_context) {
  ble_cus_t *p_cus = (ble_cus_t *)p_context;

  NRF_LOG_INFO("BLE event received. Event type = %d\r\n", p_ble_evt->header.evt_id);
  if (p_cus == NULL || p_ble_evt == NULL) {
    return;
  }

  switch (p_ble_evt->header.evt_id) {
  case BLE_GAP_EVT_CONNECTED:
    on_connect(p_cus, p_ble_evt);
    break;

  case BLE_GAP_EVT_DISCONNECTED:
    on_disconnect(p_cus, p_ble_evt);
    break;

  case BLE_GATTS_EVT_WRITE:
    NRF_LOG_DEBUG("Write EVent");
    on_write(p_cus, p_ble_evt);
    break;

  case BLE_GATTS_EVT_HVC:
    NRF_LOG_DEBUG("Indicate EVent");
    start_data_transfer(p_cus);

    break;
  default:
    // No implementation needed.
    break;
  }
}

/**@brief Function used to add the Data characteristic to the tag service
 */
static uint32_t tag_add_data_characteristic(ble_cus_t *p_cus, const ble_cus_init_t *p_cus_init) {
  uint32_t err_code;
  ble_gatts_char_md_t char_md;
  ble_gatts_attr_md_t cccd_md;
  ble_gatts_attr_t attr_char_value;
  ble_uuid_t ble_uuid;
  ble_gatts_attr_md_t attr_md;

  // Add Custom Value characteristic
  memset(&cccd_md, 0, sizeof(cccd_md));

  //  Read  operation on cccd should be possible without authentication.
  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);

  cccd_md.write_perm = p_cus_init->custom_value_char_attr_md.cccd_write_perm;
  cccd_md.vloc = BLE_GATTS_VLOC_STACK;

  memset(&char_md, 0, sizeof(char_md));

  char_md.char_props.read = 1;
  char_md.char_props.write_wo_resp = 1;
  char_md.char_props.indicate = 1;
  char_md.p_char_user_desc = NULL;
  char_md.p_char_pf = NULL;
  char_md.p_user_desc_md = NULL;
  char_md.p_cccd_md = &cccd_md;
  char_md.p_sccd_md = NULL;

  ble_uuid.type = p_cus->uuid_type;
  ble_uuid.uuid = DATA_TRANFER_CHAR_UUID;

  memset(&attr_md, 0, sizeof(attr_md));

  attr_md.read_perm = p_cus_init->custom_value_char_attr_md.read_perm;
  attr_md.write_perm = p_cus_init->custom_value_char_attr_md.write_perm;
  attr_md.vloc = BLE_GATTS_VLOC_STACK;
  attr_md.rd_auth = 0;
  attr_md.wr_auth = 0;
  attr_md.vlen = 0;

  memset(&attr_char_value, 0, sizeof(attr_char_value));

  update_tag_data();
  attr_char_value.p_uuid = &ble_uuid;
  attr_char_value.p_attr_md = &attr_md;
  attr_char_value.init_len = strlen(tag_packet_data);
  attr_char_value.init_offs = 0;
  attr_char_value.max_len = strlen(tag_packet_data);
  attr_char_value.p_value = (uint8_t *)tag_packet_data;

  err_code = sd_ble_gatts_characteristic_add(p_cus->service_handle, &char_md,
      &attr_char_value,
      &p_cus->custom_value_handles);
  if (err_code != NRF_SUCCESS) {
    return err_code;
  }
  return NRF_SUCCESS;
}

uint32_t ble_tag_services_init(ble_cus_t *p_cus, const ble_cus_init_t *p_cus_init, tag_data *tag_data_array_p, int *current_pointer_index) {
  if (p_cus == NULL || p_cus_init == NULL) {
    return NRF_ERROR_NULL;
  }

  APP_ERROR_CHECK(sd_ble_gap_addr_get(&device_addr_));

  tag_data_array_p_ = tag_data_array_p;
  current_pointer_index_p = current_pointer_index;
  uint32_t err_code;
  ble_uuid_t ble_uuid;

  // Initialize service structure
  p_cus->evt_handler = p_cus_init->evt_handler;
  p_cus->conn_handle = BLE_CONN_HANDLE_INVALID;

  // Add Custom Service UUID
  ble_uuid128_t base_uuid = {BLE_SERVICE_UUID_BASE};
  err_code = sd_ble_uuid_vs_add(&base_uuid, &p_cus->uuid_type);
  VERIFY_SUCCESS(err_code);

  ble_uuid.type = p_cus->uuid_type;
  ble_uuid.uuid = DATA_TRANSFER_SERVICE_UUID;

  // Add the Custom Service
  err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid, &p_cus->service_handle);
  if (err_code != NRF_SUCCESS) {
    return err_code;
  }

  // Add Custom Value characteristic
  return tag_add_data_characteristic(p_cus, p_cus_init);
}