#pragma once
#ifndef GATT_SERVICE_H__
#define GATT_SERVICE_H__

/**
 * @Brief BLE service library, this will have only the GATT realted functions.
 * @Updated on 10-04-2023
 **/

// Std Libs
#include <stdbool.h>
#include <stdint.h>

// BLE Libs
#include "ble.h"
#include "ble_srv_common.h"

// NRF Log Libs
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

// Custom Libs
#include "tag_utils.h"

#define NAMESPACE_SIZE 10  // Beacon namespace size
#define INSTANCE_ID_SIZE 6 // Beacon Instance Id size
#define MAC_ADDRESS_SIZE 6 // Mac address size
#define UUID_SIZE 2        // Beacon UUID size

// This is the BASE UUID: 00000000-0000-1000-8000-00805F9B34FB
#define BLE_SERVICE_UUID_BASE \
  { 0xFB, 0x34, 0x9B, 0x5F, 0x80, 0x00, 0x00, 0x80, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }

#define DATA_TRANSFER_SERVICE_UUID 0X010B
#define DATA_TRANFER_CHAR_UUID 0x1401
#define CONTACT_EPOCH_CHAR_UUID 0x1402

#define DATA_PACKET_LENGTH NRF_SDH_BLE_GATT_MAX_MTU_SIZE

#define BLE_CUS_DEF(_name)          \
  static ble_cus_t _name;           \
  NRF_SDH_BLE_OBSERVER(_name##_obs, \
      BLE_HRS_BLE_OBSERVER_PRIO,    \
      ble_events_handler, &_name)

/**
 * Data Strucuture to capture the tag data
 **/
typedef struct tag_data_t {
  uint16_t battery_percentage;
  uint8_t rssi;
  uint8_t address[MAC_ADDRESS_SIZE];
  uint64_t contact_mS;
} tag_data;

typedef struct ble_cus_s ble_cus_t;

typedef void (*ble_evt_handler_t)(ble_cus_t *p_cus_service, ble_evt_t const *p_evt);

/**@brief Battery Service init structure. This contains all options and data needed for
 *        initialization of the service.*/

typedef struct
{
  ble_evt_handler_t evt_handler;                          /**< Event handler to be called for handling events in the Custom Service. */
  uint8_t initial_custom_value;                           /**< Initial custom value */
  ble_srv_cccd_security_mode_t custom_value_char_attr_md; /**< Initial security level for Custom characteristics attribute */
} ble_cus_init_t;

/**@brief Custom Service structure. This contains various status information for the service. */
struct ble_cus_s {
  ble_evt_handler_t evt_handler;                 /**< Event handler to be called for handling events in the Custom Service. */
  uint16_t service_handle;                       /**< Handle of Custom Service (as provided by the BLE stack). */
  ble_gatts_char_handles_t custom_value_handles; /**< Handles related to the Custom Value characteristic. */
  uint16_t conn_handle;                          /**< Handle of the current connection (as provided by the BLE stack, is BLE_CONN_HANDLE_INVALID if not in a connection). */
  uint8_t uuid_type;
};

/**
 * Ble CHAR write event handler
 **/
typedef void (*on_write_handler_t)(ble_cus_t *p_cus, ble_evt_t const *p_ble_evt);

/**
 * Ble CHAR indicate event handler
 **/
typedef void (*on_indicate_handler_t)(ble_cus_t *p_cus, ble_evt_t const *p_ble_evt);

/**
 * Ble CHAR Read event handler
 **/
typedef void (*on_read_handler_t)(ble_cus_t *p_cus, ble_evt_t const *p_ble_evt);


/**
 * @brief Function for intializing the tag GATT services
 */
uint32_t gatt_services_init(ble_cus_t *p_cus, const ble_cus_init_t *p_cus_init, on_write_handler_t write_handler, on_indicate_handler_t indicate_handler,on_read_handler_t read_handler);

/**
 * @Brief Function to stop the GATT service
 */
void stop_gatt_service();

/**
 * @Breif This Function is used to send the data to the hub
 **/
uint32_t send_data_to_hub(ble_cus_t *p_cus, char *data);

void ble_events_handler(ble_evt_t const *p_ble_evt, void *p_context);

bool is_client_connected();

#endif // GATT_SERVICE_H__