/**
 * @Brief Main file for the TAG.
 * This will have the Connection/disconnection callback.
 * This will have GATT server advertising callbacks.
 * @Updated on 10-04-2023
 **/

#include "ble_component.h"

static ble_gap_addr_t tag_mac_addr_;

static uint32_t tag_start_epoch_;

/**@brief Function for handling the idle state (main loop). If there is no pending log operation,
          then sleep until next the next event occurs.
 */
static void idle_state_handle(void) {
  if (NRF_LOG_PROCESS() == false) {
    nrf_pwr_mgmt_run();
  } // end if
}

/**@brief Function for handling BLE events.
 */
static void custom_ble_evt_handler(ble_cus_t *p_cus_service, ble_evt_t const *p_ble_evt) {
  ret_code_t err_code;
  switch (p_ble_evt->header.evt_id) {
  case BLE_GAP_EVT_CONNECTED:
    NRF_LOG_DEBUG("MAIN-- HUB Connected.");
    break;

  case BLE_GAP_EVT_DISCONNECTED: {
    NRF_LOG_DEBUG("MAIN-- HUB Disconnected.");
    break;
  }
  case BLE_GAP_EVT_PHY_UPDATE_REQUEST: {
    NRF_LOG_DEBUG("MAIN-- PHY update request.");
    ble_gap_phys_t const phys =
        {
            .rx_phys = BLE_GAP_PHY_AUTO,
            .tx_phys = BLE_GAP_PHY_AUTO,
        };
    err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &phys);
    APP_ERROR_CHECK(err_code);
  } break;

  case BLE_GATTS_EVT_SYS_ATTR_MISSING:
    NRF_LOG_DEBUG("MAIN-- BLE_GATTS_EVT_SYS_ATTR_MISSING.");
    // No system attributes have been stored.
    err_code = sd_ble_gatts_sys_attr_set(p_ble_evt->evt.gatts_evt.conn_handle, NULL, 0, 0);
    APP_ERROR_CHECK(err_code);
    break;
  case BLE_GATTS_EVT_TIMEOUT:
    NRF_LOG_DEBUG("MAIN-- GATTS TIMEOUT");
    err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    APP_ERROR_CHECK(err_code);
    break;
  default:
    NRF_LOG_DEBUG("MAIN-- Invalid request %d", p_ble_evt->header.evt_id);
    break;
  }
}

static void gatts_adv_events(ble_adv_evt_t ble_adv_evt) {
  switch (ble_adv_evt) {
  case BLE_ADV_EVT_FAST:
    NRF_LOG_INFO("MAIN-- Fast advertising.");
#if TAG_DEBUG_MODE
    switch_led_debug_mode(TAG_GATT_ADV_MODE);
#endif
    break;
  case BLE_ADV_EVT_IDLE: {
    NRF_LOG_INFO("MAIN-- Advertising stopped.");
    stop_gatt_server();
    start_hub_beacon_scan();
    break;
  }
  case BLE_GAP_EVT_TIMEOUT:
    NRF_LOG_INFO("MAIN-- Advertising Timeout.");
    break;
  default:
    break;
  }
}

/**@brief Function for application main entry.
 */
int main(void) {
  NRF_LOG_INFO("*********TAG device of version %s booting up with records capacity %d*********", TAG_FIRMWARE_VER, MAX_DATA_QUEUE_LEN);

  APP_ERROR_CHECK(NRF_LOG_INIT(NULL));
  NRF_LOG_DEFAULT_BACKENDS_INIT();

  tag_init_led();
  tag_btns_init();
  tag_init_battery_pins();
  boot_timer_init();
  APP_ERROR_CHECK(nrf_pwr_mgmt_init());
  hub_beacon_scan_init();
  nrf_delay_ms(1000);
  APP_ERROR_CHECK(sd_ble_gap_addr_get(&tag_mac_addr_));
  gatt_module_init(custom_ble_evt_handler, gatts_adv_events, &tag_mac_addr_, &tag_start_epoch_);
  start_hub_beacon_scan();

  for (;;) {
    NRF_LOG_FLUSH();
    idle_state_handle();
    switch_led_debug_mode(TAG_ALL_LED_OFF);

    check_and_update_battery_state();
    if (get_scan_finshed_epoch() > 0 && abs(millis() - get_scan_finshed_epoch()) > 5000 && !is_data_transfer_in_progress()) {
      start_hub_beacon_scan();
    } // endif
  }
}